@@ -1,6 +1,7 @@
 /* GObject introspection: Repository implementation
  *
  * Copyright (C) 2005 Matthias Clasen
+ * Copyright (C) 2008,2009 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -29,21 +30,28 @@
 
 struct _GIBaseInfo 
 {
+  /* Keep this part in sync with GIUnresolvedInfo below */
   gint type;
   gint ref_count;
   GIRepository *repository;
   GIBaseInfo *container;
 
+  /* Resolved specific */
+
   GTypelib *typelib;
   guint32 offset;
 };
 
 struct _GIUnresolvedInfo
 {
+  /* Keep this part in sync with GIBaseInfo above */
   gint type;
   gint ref_count;
+  GIRepository *repository;
   GIBaseInfo *container;
 
+  /* Unresolved specific */
+
   const gchar *name;
   const gchar *namespace;
 };
@@ -195,6 +203,7 @@ g_info_from_entry (GIRepository *repository,
 
 	  unresolved->type = GI_INFO_TYPE_UNRESOLVED;
 	  unresolved->ref_count = 1;
+	  unresolved->repository = g_object_ref (repository);
 	  unresolved->container = NULL;
 	  unresolved->name = name;
 	  unresolved->namespace = namespace;
@@ -227,7 +236,8 @@ g_base_info_unref (GIBaseInfo *info)
       if (info->container)
 	g_base_info_unref (info->container);
 
-      g_object_unref (info->repository);
+      if (info->repository)
+        g_object_unref (info->repository);
 
       g_free (info);
     }
@@ -401,65 +411,131 @@ g_base_info_is_deprecated (GIBaseInfo *info)
   return FALSE;
 }
 
-static int
-cmp_annotation (const void *av,
-		const void *bv)
+/**
+ * g_base_info_get_attribute:
+ * @info: A #GIBaseInfo
+ * @name: A freeform string naming an attribute
+ *
+ * Retrieve an arbitrary attribute associated with this node.
+ *
+ * Return value: The value of the attribute, or %NULL if no such attribute exists
+ */
+const gchar *
+g_base_info_get_attribute (GIBaseInfo   *info,
+			   const gchar *name)
 {
-  const AnnotationBlob *a = av;
-  const AnnotationBlob *b = bv;
- 
-  if (b->offset < a->offset)
-    return -1;
+  GIAttributeIter iter = { 0, };
+  gchar *curname, *curvalue;
+  while (g_base_info_iterate_attributes (info, &iter, &curname, &curvalue))
+    {
+      if (strcmp (name, curname) == 0)
+        return (const gchar*) curvalue;
+    }
 
-  if (b->offset > a->offset)
-    return 1;
-  
-  return 0;
+  return NULL;
 }
 
-const gchar *
-g_base_info_get_annotation (GIBaseInfo   *info,
-			    const gchar *name)
+static int
+cmp_attribute (const void *av,
+                const void *bv)
+{
+  const AttributeBlob *a = av;
+  const AttributeBlob *b = bv;
+ 
+  if (a->offset < b->offset)
+    return -1;
+  else if (a->offset == b->offset)
+    return 0;
+  else
+    return 1;
+}
+
+static AttributeBlob *
+find_first_attribute (GIBaseInfo *info)
 {
   GIBaseInfo *base = (GIBaseInfo *)info;
   Header *header = (Header *)base->typelib->data;
-  AnnotationBlob blob, *first, *after, *res, *next;
-  const gchar *rname;
+  AttributeBlob blob, *first, *res, *previous;
 
   blob.offset = base->offset;
   
-  first = (AnnotationBlob *) &base->typelib->data[header->annotations];
-  after = (AnnotationBlob *) &base->typelib->data[header->annotations + 
-					     header->n_annotations * header->annotation_blob_size];
+  first = (AttributeBlob *) &base->typelib->data[header->attributes];
+
+  res = bsearch (&blob, first, header->n_attributes,
+                 header->attribute_blob_size, cmp_attribute);
 
-  res = bsearch (&blob, first, header->n_annotations,
-		 header->annotation_blob_size, cmp_annotation);
-  
   if (res == NULL)
     return NULL;
 
-  next = res;
-  do 
+  previous = res - 1;
+  while (previous >= first && previous->offset == base->offset)
     {
-      res = next;
-      next = res -= header->annotation_blob_size;
+      res = previous;
+      previous = res - 1;
     }
-  while (next >= first && next->offset == base->offset);
-    
-  next = res;
-  do 
-    {
-      res = next;
-      
-      rname = g_typelib_get_string (base->typelib, res->name);
-      if (strcmp (name, rname) == 0)
-	return g_typelib_get_string (base->typelib, res->value);
 
-      next = res += header->annotation_blob_size;
-    }
-  while (next < after && next->offset == base->offset);
+  return res;
+}
 
-  return NULL;
+/**
+ * g_base_info_iterate_attributes:
+ * @info: A #GIBaseInfo
+ * @iter: A #GIAttributeIter structure, must be initialized; see below
+ * @name: (out) (transfer none): Returned name, must not be freed
+ * @value: (out) (transfer none): Returned name, must not be freed
+ *
+ * Iterate over all attributes associated with this node.  The iterator
+ * structure is typically stack allocated, and must have its first
+ * member initialized to %NULL.
+ *
+ * Both the @name and @value should be treated as constants
+ * and must not be freed.
+ *
+ * <example>
+ * <title>Iterating over attributes</title>
+ * <programlisting>
+ * void
+ * print_attributes (GIBaseInfo *info)
+ * {
+ *   GIAttributeIter iter = { 0, };
+ *   char *name;
+ *   char *value;
+ *   while (g_base_info_iterate_attributes (info, &iter, &name, &value))
+ *     {
+ *       g_print ("attribute name: %s value: %s", name, value);
+ *     }
+ * }
+ * </programlisting>
+ * </example>
+ *
+ * Return value: %TRUE if there are more attributes, %FALSE otherwise
+ */
+gboolean
+g_base_info_iterate_attributes (GIBaseInfo       *info,
+                                 GIAttributeIter *iter,
+                                 gchar           **name,
+                                 gchar           **value)
+{
+  GIBaseInfo *base = (GIBaseInfo *)info;
+  Header *header = (Header *)base->typelib->data;
+  AttributeBlob *next, *after;
+
+  after = (AttributeBlob *) &base->typelib->data[header->attributes +
+                                                  header->n_attributes * header->attribute_blob_size];
+
+  if (iter->data != NULL)
+    next = (AttributeBlob *) iter->data;
+  else
+    next = find_first_attribute (info);
+
+  if (next == NULL || next->offset != base->offset || next >= after)
+    return FALSE;
+
+  *name = (gchar*) g_typelib_get_string (base->typelib, next->name);
+  *value = (gchar*) g_typelib_get_string (base->typelib, next->value);
+  iter->data = next + 1;
+
+  return TRUE;
 }
 
 GIBaseInfo *
@@ -569,7 +645,7 @@ g_type_info_new (GIBaseInfo    *container,
   SimpleTypeBlob *type = (SimpleTypeBlob *)&typelib->data[offset];
 
   return (GITypeInfo *) g_info_new (GI_INFO_TYPE_TYPE, container, typelib, 
-				    (type->reserved == 0 && type->reserved2 == 0) ? offset : type->offset);
+				    (type->flags.reserved == 0 && type->flags.reserved2 == 0) ? offset : type->offset);
 }
 
 /**
@@ -657,6 +733,7 @@ g_callable_info_get_n_args (GICallableInfo *info)
 /**
  * g_callable_info_get_arg:
  * @info: a #GICallableInfo
+ * @n: the argument index to fetch
  *
  * Get information about a particular argument of this callable.
  *
@@ -773,7 +850,7 @@ g_arg_info_get_type (GIArgInfo *info)
 {
   GIBaseInfo *base = (GIBaseInfo *)info;
  
-  return g_type_info_new (base, base->typelib, base->offset + 12);
+  return g_type_info_new (base, base->typelib, base->offset + G_STRUCT_OFFSET (ArgBlob, arg_type));
 }
 
 /* GITypeInfo functions */
@@ -783,8 +860,8 @@ g_type_info_is_pointer (GITypeInfo *info)
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (type->reserved == 0 && type->reserved2 == 0)
-    return type->pointer;
+  if (type->flags.reserved == 0 && type->flags.reserved2 == 0)
+    return type->flags.pointer;
   else
     {
       InterfaceTypeBlob *iface = (InterfaceTypeBlob *)&base->typelib->data[base->offset];
@@ -799,8 +876,8 @@ g_type_info_get_tag (GITypeInfo *info)
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
 
-  if (type->reserved == 0 && type->reserved2 == 0)
-    return type->tag;
+  if (type->flags.reserved == 0 && type->flags.reserved2 == 0)
+    return type->flags.tag;
   else
     {
       InterfaceTypeBlob *iface = (InterfaceTypeBlob *)&base->typelib->data[base->offset];
@@ -816,7 +893,7 @@ g_type_info_get_param_type (GITypeInfo *info,
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (!(type->reserved == 0 && type->reserved2 == 0))
+  if (!(type->flags.reserved == 0 && type->flags.reserved2 == 0))
     {
       ParamTypeBlob *param = (ParamTypeBlob *)&base->typelib->data[base->offset];
 
@@ -826,7 +903,9 @@ g_type_info_get_param_type (GITypeInfo *info,
 	case GI_TYPE_TAG_GLIST:
 	case GI_TYPE_TAG_GSLIST:
 	case GI_TYPE_TAG_GHASH:
-	  return g_type_info_new (base, base->typelib, base->offset + 4 + 4 * n);
+	  return g_type_info_new (base, base->typelib,
+	                          base->offset + sizeof (ParamTypeBlob)
+	                          + sizeof (SimpleTypeBlob) * n);
 	  break;
 	  
 	default: ;
@@ -842,7 +921,7 @@ g_type_info_get_interface (GITypeInfo *info)
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (!(type->reserved == 0 && type->reserved2 == 0))
+  if (!(type->flags.reserved == 0 && type->flags.reserved2 == 0))
     {
       InterfaceTypeBlob *blob = (InterfaceTypeBlob *)&base->typelib->data[base->offset];
       
@@ -859,14 +938,14 @@ g_type_info_get_array_length (GITypeInfo *info)
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (!(type->reserved == 0 && type->reserved2 == 0))
+  if (!(type->flags.reserved == 0 && type->flags.reserved2 == 0))
     {
       ArrayTypeBlob *blob = (ArrayTypeBlob *)&base->typelib->data[base->offset];
 
       if (blob->tag == GI_TYPE_TAG_ARRAY)
 	{
 	  if (blob->has_length)
-	    return blob->length;
+	    return blob->dimensions.length;
 	}
     }
 
@@ -879,14 +958,14 @@ g_type_info_get_array_fixed_size (GITypeInfo *info)
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (!(type->reserved == 0 && type->reserved2 == 0))
+  if (!(type->flags.reserved == 0 && type->flags.reserved2 == 0))
     {
       ArrayTypeBlob *blob = (ArrayTypeBlob *)&base->typelib->data[base->offset];
 
       if (blob->tag == GI_TYPE_TAG_ARRAY)
 	{
 	  if (blob->has_size)
-	    return blob->size;
+	    return blob->dimensions.size;
 	}
     }
 
@@ -899,7 +978,7 @@ g_type_info_is_zero_terminated (GITypeInfo *info)
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (!(type->reserved == 0 && type->reserved2 == 0))
+  if (!(type->flags.reserved == 0 && type->flags.reserved2 == 0))
     {
       ArrayTypeBlob *blob = (ArrayTypeBlob *)&base->typelib->data[base->offset];
 
@@ -916,7 +995,7 @@ g_type_info_get_n_error_domains (GITypeInfo *info)
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (!(type->reserved == 0 && type->reserved2 == 0))
+  if (!(type->flags.reserved == 0 && type->flags.reserved2 == 0))
     {
       ErrorTypeBlob *blob = (ErrorTypeBlob *)&base->typelib->data[base->offset];
 
@@ -934,7 +1013,7 @@ g_type_info_get_error_domain (GITypeInfo *info,
   GIBaseInfo *base = (GIBaseInfo *)info;
   SimpleTypeBlob *type = (SimpleTypeBlob *)&base->typelib->data[base->offset];
   
-  if (!(type->reserved == 0 && type->reserved2 == 0))
+  if (!(type->flags.reserved == 0 && type->flags.reserved2 == 0))
     {
       ErrorTypeBlob *blob = (ErrorTypeBlob *)&base->typelib->data[base->offset];
 
@@ -1022,7 +1101,7 @@ g_field_info_get_type (GIFieldInfo *info)
 {
   GIBaseInfo *base = (GIBaseInfo *)info;
   
-  return g_type_info_new (base, base->typelib, base->offset + 8);
+  return g_type_info_new (base, base->typelib, base->offset + G_STRUCT_OFFSET (FieldBlob, type));
 }
 
 /* GIRegisteredTypeInfo functions */
@@ -1060,6 +1139,8 @@ g_registered_type_info_get_g_type (GIRegisteredTypeInfo *info)
   
   if (type_init == NULL)
     return G_TYPE_NONE;
+  else if (!strcmp (type_init, "intern"))
+    return G_TYPE_OBJECT;
   
   get_type_func = NULL;
   if (!g_typelib_symbol (((GIBaseInfo*)info)->typelib,
@@ -1175,6 +1256,25 @@ g_struct_info_get_alignment (GIStructInfo *info)
   return blob->alignment;
 }
 
+/**
+ * g_struct_info_is_gtype_struct:
+ * @info: GIStructInfo
+ * 
+ * Return true if this structure represents the "class structure" for some
+ * #GObject or #GInterface.  This function is mainly useful to hide this kind of structure
+ * from generated public APIs.
+ *
+ * Returns: %TRUE if this is a class struct, %FALSE otherwise
+ */
+gboolean
+g_struct_info_is_gtype_struct (GIStructInfo *info)
+{
+  GIBaseInfo *base = (GIBaseInfo *)info;
+  StructBlob *blob = (StructBlob *)&base->typelib->data[base->offset];
+
+  return blob->is_gtype_struct;
+}
+
 gint
 g_enum_info_get_n_values (GIEnumInfo *info)
 {
@@ -1435,6 +1535,63 @@ g_object_info_get_vfunc (GIObjectInfo *info,
 				     base->typelib, offset);  
 }
 
+static GIVFuncInfo *
+find_vfunc (GIBaseInfo   *base,
+            guint32       offset,
+            gint          n_vfuncs,
+            const gchar  *name)
+{
+  /* FIXME hash */
+  Header *header = (Header *)base->typelib->data;
+  gint i;
+
+  for (i = 0; i < n_vfuncs; i++)
+    {
+      VFuncBlob *fblob = (VFuncBlob *)&base->typelib->data[offset];
+      const gchar *fname = (const gchar *)&base->typelib->data[fblob->name];
+
+      if (strcmp (name, fname) == 0)
+        return (GIVFuncInfo *) g_info_new (GI_INFO_TYPE_VFUNC, base,
+                                           base->typelib, offset);
+
+      offset += header->vfunc_blob_size;
+    }
+
+  return NULL;
+}
+
+/**
+ * g_object_info_find_vfunc:
+ * @info: An #GIObjectInfo
+ * @name: The name of a virtual function to find.
+ *
+ * Locate a virtual function slot with name @name.  Note that the namespace
+ * for virtuals is distinct from that of methods; there may or may not be
+ * a concrete method associated for a virtual.  If there is one, it may
+ * be retrieved using #g_vfunc_info_get_invoker.  See the documentation for
+ * that function for more information on invoking virtuals.
+ *
+ * Return value: (transfer full): A #GIVFuncInfo, or %NULL if none with name @name.
+ */
+GIVFuncInfo *
+g_object_info_find_vfunc (GIObjectInfo *info,
+                          const gchar  *name)
+{
+  gint offset;
+  GIBaseInfo *base = (GIBaseInfo *)info;
+  Header *header = (Header *)base->typelib->data;
+  ObjectBlob *blob = (ObjectBlob *)&base->typelib->data[base->offset];
+
+  offset = base->offset + header->object_blob_size
+    + (blob->n_interfaces + blob->n_interfaces % 2) * 2
+    + blob->n_fields * header->field_blob_size
+    + blob->n_properties * header->property_blob_size
+    + blob->n_methods * header->function_blob_size
+    + blob->n_signals * header->signal_blob_size;
+
+  return find_vfunc (base, offset, blob->n_vfuncs, name);
+}
+
 gint
 g_object_info_get_n_constants (GIObjectInfo *info)
 {
@@ -1466,6 +1623,27 @@ g_object_info_get_constant (GIObjectInfo *info,
 					base->typelib, offset);  
 }
 
+/**
+ * g_object_info_get_class_struct:
+ * @info: A #GIObjectInfo to query
+ * 
+ * Every #GObject has two structures; an instance structure and a class
+ * structure.  This function returns the metadata for the class structure.
+ *
+ * Returns: a #GIStructInfo for the class struct or %NULL if none found.
+ */
+GIStructInfo *
+g_object_info_get_class_struct (GIObjectInfo *info)
+{
+  GIBaseInfo *base = (GIBaseInfo *)info;
+  ObjectBlob *blob = (ObjectBlob *)&base->typelib->data[base->offset];
+
+  if (blob->gtype_struct)
+    return (GIStructInfo *) g_info_from_entry (base->repository,
+                                               base->typelib, blob->gtype_struct);
+  else
+    return NULL;
+}
 
 /* GIInterfaceInfo functions */
 gint
@@ -1615,6 +1793,34 @@ g_interface_info_get_vfunc (GIInterfaceInfo *info,
 				     base->typelib, offset);  
 }
 
+/**
+ * g_interface_info_find_vfunc:
+ * @info: An #GIObjectInfo
+ * @name: The name of a virtual function to find.
+ *
+ * Locate a virtual function slot with name @name.  See the documentation
+ * for #g_object_info_find_vfunc for more information on virtuals.
+ *
+ * Return value: (transfer full): A #GIVFuncInfo, or %NULL if none with name @name.
+ */
+GIVFuncInfo *
+g_interface_info_find_vfunc (GIInterfaceInfo *info,
+                             const gchar  *name)
+{
+  gint offset;
+  GIBaseInfo *base = (GIBaseInfo *)info;
+  Header *header = (Header *)base->typelib->data;
+  InterfaceBlob *blob = (InterfaceBlob *)&base->typelib->data[base->offset];
+
+  offset = base->offset + header->interface_blob_size
+    + (blob->n_prerequisites + blob->n_prerequisites % 2) * 2
+    + blob->n_properties * header->property_blob_size
+    + blob->n_methods * header->function_blob_size
+    + blob->n_signals * header->signal_blob_size;
+
+  return find_vfunc (base, offset, blob->n_vfuncs, name);
+}
+
 gint
 g_interface_info_get_n_constants (GIInterfaceInfo *info)
 {
@@ -1645,8 +1851,26 @@ g_interface_info_get_constant (GIInterfaceInfo *info,
 					base->typelib, offset);  
 }
 
+/**
+ * g_interface_info_get_iface_struct:
+ * @info: A #GIInterfaceInfo to query
+ *
+ * Returns the layout C structure associated with this #GInterface.
+ *
+ * Returns: A #GIStructInfo for the class struct or %NULL if none found.
+ */
+GIStructInfo *
+g_interface_info_get_iface_struct (GIInterfaceInfo *info)
+{
+  GIBaseInfo *base = (GIBaseInfo *)info;
+  InterfaceBlob *blob = (InterfaceBlob *)&base->typelib->data[base->offset];
 
-
+  if (blob->gtype_struct)
+    return (GIStructInfo *) g_info_from_entry (base->repository,
+                                               base->typelib, blob->gtype_struct);
+  else
+    return NULL;
+}
 
 /* GIPropertyInfo functions */
 GParamFlags
@@ -1678,7 +1902,7 @@ g_property_info_get_type (GIPropertyInfo *info)
 {
   GIBaseInfo *base = (GIBaseInfo *)info;
 
-  return g_type_info_new (base, base->typelib, base->offset + 8);  
+  return g_type_info_new (base, base->typelib, base->offset + G_STRUCT_OFFSET (PropertyBlob, type));
 }
 
 
@@ -1782,6 +2006,37 @@ g_vfunc_info_get_signal (GIVFuncInfo *info)
   return NULL;
 }
 
+/**
+ * g_vfunc_info_get_invoker:
+ * @info: A #GIVFuncInfo
+ *
+ * If this virtual function has an associated invoker method, this
+ * method will return it.  An invoker method is a C entry point.
+ *
+ * Not all virtuals will have invokers.
+ *
+ * Return value: (transfer full): An invoker function, or %NULL if none known
+ */
+GIFunctionInfo *
+g_vfunc_info_get_invoker (GIVFuncInfo *info)
+{
+  GIBaseInfo *base = (GIBaseInfo *)info;
+  VFuncBlob *blob = (VFuncBlob *)&base->typelib->data[base->offset];
+  GIBaseInfo *container = base->container;
+  GIInfoType parent_type;
+
+  /* 1023 = 0x3ff is the maximum of the 10 bits for invoker index */
+  if (blob->invoker == 1023)
+    return NULL;
+
+  parent_type = g_base_info_get_type (container);
+  if (parent_type == GI_INFO_TYPE_OBJECT)
+    return g_object_info_get_method ((GIObjectInfo*)container, blob->invoker);
+  else if (parent_type == GI_INFO_TYPE_INTERFACE)
+    return g_interface_info_get_method ((GIInterfaceInfo*)container, blob->invoker);
+  else
+    g_assert_not_reached ();
+}
 
 /* GIConstantInfo functions */
 GITypeInfo *
@@ -1800,13 +2055,13 @@ g_constant_info_get_value (GIConstantInfo *info,
   ConstantBlob *blob = (ConstantBlob *)&base->typelib->data[base->offset];
 
   /* FIXME non-basic types ? */
-  if (blob->type.reserved == 0 && blob->type.reserved2 == 0)
+  if (blob->type.flags.reserved == 0 && blob->type.flags.reserved2 == 0)
     {
-      if (blob->type.pointer)
+      if (blob->type.flags.pointer)
 	value->v_pointer = g_memdup (&base->typelib->data[blob->offset], blob->size);
       else
 	{
-	  switch (blob->type.tag)
+	  switch (blob->type.flags.tag)
 	    {
 	    case GI_TYPE_TAG_BOOLEAN:
 	      value->v_boolean = *(gboolean*)&base->typelib->data[blob->offset];
@@ -1844,6 +2099,12 @@ g_constant_info_get_value (GIConstantInfo *info,
 	    case GI_TYPE_TAG_TIME_T:
 	      value->v_long = *(long*)&base->typelib->data[blob->offset];
 	      break;
+	    case GI_TYPE_TAG_SHORT:
+	      value->v_short = *(gshort*)&base->typelib->data[blob->offset];
+	      break;
+	    case GI_TYPE_TAG_USHORT:
+	      value->v_ushort = *(gushort*)&base->typelib->data[blob->offset];
+	      break;
 	    case GI_TYPE_TAG_INT:
 	      value->v_int = *(gint*)&base->typelib->data[blob->offset];
 	      break;
