@@ -37,11 +37,13 @@
 static GStaticMutex globals_lock = G_STATIC_MUTEX_INIT;
 static GIRepository *default_repository = NULL;
 static GSList *search_path = NULL;
+static GSList *override_search_path = NULL;
 
 struct _GIRepositoryPrivate 
 {
   GHashTable *typelibs; /* (string) namespace -> GTypelib */
   GHashTable *lazy_typelibs; /* (string) namespace-version -> GTypelib */
+  GHashTable *info_by_gtype; /* GType -> GIBaseInfo */
 };
 
 G_DEFINE_TYPE (GIRepository, g_irepository, G_TYPE_OBJECT);
@@ -57,6 +59,10 @@ g_irepository_init (GIRepository *repository)
 			     (GDestroyNotify) g_typelib_free);
   repository->priv->lazy_typelibs 
     = g_hash_table_new (g_str_hash, g_str_equal);
+  repository->priv->info_by_gtype
+    = g_hash_table_new_full (g_direct_hash, g_direct_equal,
+                             (GDestroyNotify) NULL,
+                             (GDestroyNotify) g_base_info_unref);
 }
 
 static void
@@ -66,6 +72,7 @@ g_irepository_finalize (GObject *object)
 
   g_hash_table_destroy (repository->priv->typelibs);
   g_hash_table_destroy (repository->priv->lazy_typelibs);
+  g_hash_table_destroy (repository->priv->info_by_gtype);
   
   (* G_OBJECT_CLASS (g_irepository_parent_class)->finalize) (G_OBJECT (repository));
 }
@@ -98,9 +105,13 @@ init_globals (void)
       char *typelib_dir;
       const gchar *type_lib_path_env;
 
+      /* This variable is intended to take precedence over both the default
+       * search path, as well as anything written into code with g_irepository_prepend_search_path.
+       */
       type_lib_path_env = g_getenv ("GI_TYPELIB_PATH");
 
       search_path = NULL;
+      override_search_path = NULL;
       if (type_lib_path_env)
         {
           gchar **custom_dirs;
@@ -111,7 +122,7 @@ init_globals (void)
           d = custom_dirs;
           while (*d)
             {
-              search_path = g_slist_prepend (search_path, *d);
+              override_search_path = g_slist_prepend (override_search_path, *d);
               d++;
             }
 
@@ -121,7 +132,7 @@ init_globals (void)
 
       libdir = GOBJECT_INTROSPECTION_LIBDIR;
 
-      typelib_dir = g_build_filename (libdir, "girepository", NULL);
+      typelib_dir = g_build_filename (libdir, "girepository-1.0", NULL);
 
       search_path = g_slist_prepend (search_path, typelib_dir);
 
@@ -153,6 +164,21 @@ g_irepository_get_search_path (void)
   return search_path;
 }
 
+static
+GSList *
+build_search_path_with_overrides (void)
+{
+	GSList *result;
+	if (override_search_path != NULL) 
+	  {
+		result = g_slist_copy (override_search_path);
+		g_slist_last (result)->next = g_slist_copy (search_path);
+	  } 
+	else
+	  result = g_slist_copy (search_path);
+	return result;
+}
+
 static char *
 build_typelib_key (const char *name, const char *source)
 {
@@ -337,19 +363,16 @@ register_internal (GIRepository *repository,
       g_hash_table_insert (repository->priv->typelibs, key, (void *)typelib);
     }
 
-  if (typelib->modules == NULL)
-    typelib->modules = g_list_append(typelib->modules, g_module_open (NULL, 0));
-
   return namespace;
 }
 
 /**
- * g_irepository_get_dependencies
+ * g_irepository_get_dependencies:
  * @repository: A #GIRepository, may be %NULL for the default
- * @namespace: Namespace of interest
+ * @namespace_: Namespace of interest
  *
  * Return an array of all (transitive) dependencies for namespace
- * @namespace, including version.  The returned strings are of the
+ * @namespace_, including version.  The returned strings are of the
  * form <code>namespace-version</code>.
  *
  * Note: The namespace must have already been loaded using a function
@@ -410,10 +433,10 @@ g_irepository_load_typelib (GIRepository *repository,
 }
 
 /**
- * g_irepository_is_registered
+ * g_irepository_is_registered:
  * @repository: A #GIRepository, may be %NULL for the default
- * @namespace: Namespace of interest
- * @version: <allow-none>: Required version, may be %NULL for latest
+ * @namespace_: Namespace of interest
+ * @version: (allow-none): Required version, may be %NULL for latest
  *
  * Check whether a particular namespace (and optionally, a specific
  * version thereof) is currently loaded.  This function is likely to
@@ -434,7 +457,7 @@ g_irepository_is_registered (GIRepository *repository,
 }
 
 /**
- * g_irepository_get_default
+ * g_irepository_get_default:
  *
  * Returns the singleton process-global default #GIRepository.  It is
  * not currently supported to have multiple repositories in a
@@ -456,12 +479,12 @@ g_irepository_get_default (void)
 }
 
 /**
- * g_irepository_get_n_infos
+ * g_irepository_get_n_infos:
  * @repository: A #GIRepository, may be %NULL for the default
- * @namespace: Namespace to inspect
+ * @namespace_: Namespace to inspect
  *
  * This function returns the number of metadata entries in
- * given namespace @namespace.  The namespace must have
+ * given namespace @namespace_.  The namespace must have
  * already been loaded before calling this function.
  *
  * Returns: number of metadata entries
@@ -491,6 +514,7 @@ typedef struct
   GIRepository *repo;
   gint index;
   const gchar *name;
+  gboolean type_firstpass;
   const gchar *type;
   GIBaseInfo *iface;
 } IfaceData;
@@ -502,6 +526,7 @@ find_interface (gpointer key,
 {
   gint i;
   GTypelib *typelib = (GTypelib *)value;
+  Header *header = (Header *) typelib->data;
   IfaceData *iface_data = (IfaceData *)data;
   gint index;
   gint n_entries;
@@ -527,6 +552,28 @@ find_interface (gpointer key,
     }
   else if (iface_data->type)
     {
+      const char *c_prefix;
+      /* Inside each typelib, we include the "C prefix" which acts as
+       * a namespace mechanism.  For GtkTreeView, the C prefix is Gtk.
+       * Given the assumption that GTypes for a library also use the
+       * C prefix, we know we can skip examining a typelib if our
+       * target type does not have this typelib's C prefix.
+       *
+       * However, not every class library necessarily conforms to this,
+       * e.g. Clutter has Cogl inside it.  So, we split this into two
+       * passes.  First we try a lookup, skipping things which don't
+       * have the prefix.  If that fails then we try a global lookup,
+       * ignoring the prefix.
+       *
+       * See http://bugzilla.gnome.org/show_bug.cgi?id=564016
+       */
+      c_prefix = g_typelib_get_string (typelib, header->c_prefix);
+      if (iface_data->type_firstpass && c_prefix != NULL)
+        {
+          if (g_ascii_strncasecmp (c_prefix, iface_data->type, strlen (c_prefix)) != 0)
+            return;
+        }
+
       for (i = 1; i <= n_entries; i++)
 	{
 	  RegisteredTypeBlob *blob;
@@ -565,13 +612,13 @@ find_interface (gpointer key,
 }
 
 /**
- * g_irepository_get_info
+ * g_irepository_get_info:
  * @repository: A #GIRepository, may be %NULL for the default
- * @namespace: Namespace to inspect
+ * @namespace_: Namespace to inspect
  * @index: Offset into namespace metadata for entry
  *
  * This function returns a particular metadata entry in the
- * given namespace @namespace.  The namespace must have
+ * given namespace @namespace_.  The namespace must have
  * already been loaded before calling this function.
  *
  * Returns: #GIBaseInfo containing metadata
@@ -604,43 +651,63 @@ g_irepository_get_info (GIRepository *repository,
 }
 
 /**
- * g_irepository_find_by_gtype
+ * g_irepository_find_by_gtype:
  * @repository: A #GIRepository, may be %NULL for the default
- * @type: GType to search for
+ * @gtype: GType to search for
  *
  * Searches all loaded namespaces for a particular #GType.  Note that
  * in order to locate the metadata, the namespace corresponding to
  * the type must first have been loaded.  There is currently no
  * mechanism for determining the namespace which corresponds to an
- * arbitrary GType - thus, this function will function most reliably
- * when you have expect the GType to be from a known namespace.
+ * arbitrary GType - thus, this function will operate most reliably
+ * when you know the GType to originate from be from a loaded namespace.
  *
  * Returns: #GIBaseInfo representing metadata about @type, or %NULL
  */
 GIBaseInfo * 
 g_irepository_find_by_gtype (GIRepository *repository,
-			     GType         type)
+			     GType         gtype)
 {
   IfaceData data;
 
   repository = get_repository (repository);
 
+  data.iface = g_hash_table_lookup (repository->priv->info_by_gtype,
+                                    (gpointer)gtype);
+
+  if (data.iface)
+    return g_base_info_ref (data.iface);
+
   data.repo = repository;
   data.name = NULL;
-  data.type = g_type_name (type);
+  data.type_firstpass = TRUE;
+  data.type = g_type_name (gtype);
   data.index = -1;
   data.iface = NULL;
 
   g_hash_table_foreach (repository->priv->typelibs, find_interface, &data);
   g_hash_table_foreach (repository->priv->lazy_typelibs, find_interface, &data);
 
+  /* We do two passes; see comment in find_interface */
+  if (!data.iface)
+    {
+      data.type_firstpass = FALSE;
+      g_hash_table_foreach (repository->priv->typelibs, find_interface, &data);
+      g_hash_table_foreach (repository->priv->lazy_typelibs, find_interface, &data);
+    }
+
+  if (data.iface)
+    g_hash_table_insert (repository->priv->info_by_gtype,
+                         (gpointer) gtype,
+                         g_base_info_ref (data.iface));
+
   return data.iface;
 }
 
 /**
- * g_irepository_find_by_name
+ * g_irepository_find_by_name:
  * @repository: A #GIRepository, may be %NULL for the default
- * @namespace: Namespace which will be searched
+ * @namespace_: Namespace which will be searched
  * @name: Entry name to find
  *
  * Searches for a particular entry in a namespace.  Before calling
@@ -688,12 +755,12 @@ collect_namespaces (gpointer key,
 }
 
 /**
- * g_irepository_get_namespaces
+ * g_irepository_get_namespaces:
  * @repository: A #GIRepository, may be %NULL for the default
  *
  * Return the list of currently loaded namespaces.
  *
- * Returns: <utf8,transfer>: List of namespaces
+ * Returns: (utf8) (transfer full): List of namespaces
  */
 gchar ** 
 g_irepository_get_loaded_namespaces (GIRepository *repository)
@@ -717,12 +784,12 @@ g_irepository_get_loaded_namespaces (GIRepository *repository)
 }
 
 /**
- * g_irepository_get_version
+ * g_irepository_get_version:
  * @repository: A #GIRepository, may be %NULL for the default
- * @namespace: Namespace to inspect
+ * @namespace_: Namespace to inspect
  *
  * This function returns the loaded version associated with the given
- * namespace @namespace.
+ * namespace @namespace_.
  *
  * Note: The namespace must have already been loaded using a function
  * such as #g_irepository_require before calling this function.
@@ -749,12 +816,12 @@ g_irepository_get_version (GIRepository *repository,
 }
 
 /**
- * g_irepository_get_shared_library
+ * g_irepository_get_shared_library:
  * @repository: A #GIRepository, may be %NULL for the default
- * @namespace: Namespace to inspect
+ * @namespace_: Namespace to inspect
  *
  * This function returns the full path to the shared C library
- * associated with the given namespace @namespace. There may be no
+ * associated with the given namespace @namespace_. There may be no
  * shared library path associated, in which case this function will
  * return %NULL.
  *
@@ -785,18 +852,53 @@ g_irepository_get_shared_library (GIRepository *repository,
     return NULL;
 }
 
+/**
+ * g_irepository_get_c_prefix
+ * @repository: A #GIRepository, may be %NULL for the default
+ * @namespace: Namespace to inspect
+ *
+ * This function returns the "C prefix", or the C level namespace
+ * associated with the given introspection namespace.  Each C symbol
+ * starts with this prefix, as well each #GType in the library.
+ *
+ * Note: The namespace must have already been loaded using a function
+ * such as #g_irepository_require before calling this function.
+ *
+ * Returns: C namespace prefix, or %NULL if none associated
+ */
+const gchar *
+g_irepository_get_c_prefix (GIRepository *repository,
+                            const gchar  *namespace_)
+{
+  GTypelib *typelib;
+  Header *header;
+
+  g_return_val_if_fail (namespace_ != NULL, NULL);
+
+  repository = get_repository (repository);
+
+  typelib = get_registered (repository, namespace_, NULL);
+
+  g_return_val_if_fail (typelib != NULL, NULL);
+
+  header = (Header *) typelib->data;
+  if (header->shared_library)
+    return g_typelib_get_string (typelib, header->c_prefix);
+  else
+    return NULL;
+}
+
 /**
  * g_irepository_get_typelib_path
  * @repository: Repository, may be %NULL for the default
- * @namespace: GI namespace to use, e.g. "Gtk"
- * @version: <allow-none>: Version of namespace to use, e.g. "0.8", may be %NULL
+ * @namespace_: GI namespace to use, e.g. "Gtk"
  *
- * If namespace @namespace is loaded, return the full path to the
+ * If namespace @namespace_ is loaded, return the full path to the
  * .typelib file it was loaded from.  If the typelib for 
- * namespace @namespace was included in a shared library, return
- * the special string "<builtin>".
+ * namespace @namespace_ was included in a shared library, return
+ * the special string "$lt;builtin$gt;".
  *
- * Returns: Filesystem path (or <builtin>) if successful, %NULL if namespace is not loaded
+ * Returns: Filesystem path (or $lt;builtin$gt;) if successful, %NULL if namespace is not loaded
  */
 
 const gchar * 
@@ -825,6 +927,7 @@ find_namespace_version (const gchar  *namespace,
 			const gchar  *version,
 			gchar       **path_ret)
 {
+  GSList *tmp_path;
   GSList *ldir;
   GError *error = NULL;
   GMappedFile *mfile = NULL;
@@ -832,7 +935,8 @@ find_namespace_version (const gchar  *namespace,
  
   fname = g_strdup_printf ("%s-%s.typelib", namespace, version);
 
-  for (ldir = search_path; ldir; ldir = ldir->next)
+  tmp_path = build_search_path_with_overrides ();
+  for (ldir = tmp_path; ldir; ldir = ldir->next)
     {
       char *path = g_build_filename (ldir->data, fname, NULL);
       
@@ -847,6 +951,7 @@ find_namespace_version (const gchar  *namespace,
       break;
     }
   g_free (fname);
+  g_slist_free (tmp_path);
   return mfile;
 }
 
@@ -945,6 +1050,7 @@ find_namespace_latest (const gchar  *namespace,
 		       gchar       **version_ret,
 		       gchar       **path_ret)
 {
+  GSList *tmp_path;
   GSList *ldir;
   GError *error = NULL;
   char *namespace_dash;
@@ -960,7 +1066,8 @@ find_namespace_latest (const gchar  *namespace,
   namespace_typelib = g_strdup_printf ("%s.typelib", namespace);
 
   index = 0;
-  for (ldir = search_path; ldir; ldir = ldir->next)
+  tmp_path = build_search_path_with_overrides ();  
+  for (ldir = tmp_path; ldir; ldir = ldir->next)
     {
       GDir *dir;
       const char *dirname;
@@ -1030,22 +1137,22 @@ find_namespace_latest (const gchar  *namespace,
       g_slist_foreach (candidates, (GFunc) free_candidate, NULL);
       g_slist_free (candidates);
     }  
-
   g_free (namespace_dash);
   g_free (namespace_typelib);
+  g_slist_free (tmp_path);  
   return result;
 }
 
 /**
- * g_irepository_require
- * @repository: <allow-none>: Repository, may be %NULL for the default
- * @namespace: GI namespace to use, e.g. "Gtk"
- * @version: <allow-none>: Version of namespace, may be %NULL for latest
+ * g_irepository_require:
+ * @repository: (allow-none): Repository, may be %NULL for the default
+ * @namespace_: GI namespace to use, e.g. "Gtk"
+ * @version: (allow-none): Version of namespace, may be %NULL for latest
  * @flags: Set of %GIRepositoryLoadFlags, may be %0
  * @error: a #GError.
  *
- * Force the namespace @namespace to be loaded if it isn't already.
- * If @namespace is not loaded, this function will search for a
+ * Force the namespace @namespace_ to be loaded if it isn't already.
+ * If @namespace_ is not loaded, this function will search for a
  * ".typelib" file using the repository search path.  In addition, a
  * version @version of namespace may be specified.  If @version is
  * not specified, the latest will be used.
@@ -1156,7 +1263,8 @@ g_irepository_introspect_cb (const char *option_name,
 			     gpointer data,
 			     GError **error)
 {
-  return g_irepository_dump (value, error);
+  gboolean ret = g_irepository_dump (value, error);
+  exit (ret ? 0 : 1);
 }
 
 static const GOptionEntry introspection_args[] = {
@@ -1210,6 +1318,10 @@ g_type_tag_to_string (GITypeTag type)
       return "int64";
     case GI_TYPE_TAG_UINT64:
       return "uint64";
+    case GI_TYPE_TAG_SHORT:
+      return "short";
+    case GI_TYPE_TAG_USHORT:
+      return "ushort";
     case GI_TYPE_TAG_INT:
       return "int";
     case GI_TYPE_TAG_UINT:
